#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import ccxt

import trading_backend.errors


class Exchange:
    SPOT_ID = None
    MARGIN_ID = None
    FUTURE_ID = None
    IS_SPONSORING = False

    def __init__(self, exchange):
        self._exchange = exchange

        # add backend headers
        self._exchange.connector.add_headers(self.get_headers())

    @classmethod
    def get_name(cls):
        return 'default'

    @classmethod
    def is_sponsoring(cls) -> bool:
        return cls.IS_SPONSORING

    def get_headers(self) -> dict:
        return {}

    def get_orders_parameters(self, params=None) -> dict:
        if params is None:
            params = {}
        return params

    async def is_valid_account(self) -> (bool, str):
        try:
            return await self._inner_is_valid_account()
        except ccxt.InvalidNonce as err:
            raise trading_backend.errors.TimeSyncError(err)
        except ccxt.ExchangeError as err:
            raise trading_backend.errors.ExchangeAuthError(err)

    async def _inner_is_valid_account(self) -> (bool, str):
        await self._exchange.connector.client.fetch_balance()
        return True, None

    def _get_id(self):
        if self._exchange.exchange_manager.is_future:
            return self.FUTURE_ID
        if self._exchange.exchange_manager.is_margin:
            return self.MARGIN_ID
        return self.SPOT_ID



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import trading_backend.exchanges as exchanges


class OKX(exchanges.Exchange):
    SPOT_ID = "c812bf5944b749BC"
    MARGIN_ID = "c812bf5944b749BC"
    FUTURE_ID = "c812bf5944b749BC"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'okx'

    def get_orders_parameters(self, params=None) -> dict:
        if self._exchange.connector.client.options.get("brokerId", "") != self._get_id():
            self._exchange.connector.client.options["brokerId"] = self._get_id()
        return super().get_orders_parameters(params)

    async def _inner_is_valid_account(self) -> (bool, str):
        # Nothing to do
        return await super()._inner_is_valid_account()



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import aiohttp.streams

import trading_backend.exchanges as exchanges


class Binance(exchanges.Exchange):
    SPOT_ID = "T9698EB7"
    MARGIN_ID = None
    FUTURE_ID = "uquVg2pc"
    REF_ID = "135007948"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'binance'

    def _get_order_custom_id(self):
        return f"x-{self._get_id()}{self._exchange.connector.client.uuid22()}"

    def get_orders_parameters(self, params=None) -> dict:
        params = super().get_orders_parameters(params)
        params.update({'newClientOrderId': self._get_order_custom_id()})
        return params

    async def _inner_is_valid_account(self) -> (bool, str):
        details = None
        try:
            details = await self._exchange.connector.client.sapi_get_apireferral_ifnewuser(
                params=self._exchange._get_params({
                    "apiAgentCode": self._get_id()
                })
            )
            if not details.get("rebateWorking", False):
                ref_id = details.get("referrerId", None)
                if ref_id is not None:
                    return False, f"This account has a referral id equal to {ref_id} " \
                                  f"which is incompatible ({self.REF_ID} as referral id or no referral id is required)"
                return False, f"This account is incompatible, details: {details}. Please report this message to " \
                              f"admins for investigation. " \
                              f"An account with {self.REF_ID} as referral id or no referral id is required."
            if not details.get("ifNewUser", False):
                return False, "Binance requires accounts that were created after july 1st 2021, " \
                              "this account is too old."
        except AttributeError:
            if isinstance(details, aiohttp.streams.StreamReader):
                return False, "Error when fetching exchange data (unreadable response)"
            return False, "Invalid request parameters"
        return True, None



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import ccxt.async_support
import trading_backend.exchanges as exchanges
import trading_backend
from tests import binance_exchange


def test_get_name(binance_exchange):
    assert exchanges.Binance(binance_exchange).get_name() == ccxt.async_support.binance().name.lower()


def test_get_orders_parameters(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    with mock.patch.object(exchange._exchange.connector.client, "uuid22",
                           mock.Mock(return_value="123456789")) as uuid22_mock:
        assert exchange.get_orders_parameters({"a": 1}) == {
            "a": 1,
            'newClientOrderId': exchange._get_order_custom_id()
        }
        # once by get_orders_parameters and once by _get_order_custom_id
        assert uuid22_mock.call_count == 2


@pytest.mark.asyncio
async def test_is_valid_account(binance_exchange):
    exchange = exchanges.Binance(binance_exchange)
    params = {"apiAgentCode": exchange._get_id()}
    with pytest.raises(trading_backend.ExchangeAuthError):
        await exchange.is_valid_account()
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value={"rebateWorking": False})) as sapi_get_apireferral_ifnewuser_mock:
        results = await exchange.is_valid_account()
        assert results[0] is False
        assert isinstance(results[1], str)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value={"ifNewUser": False})) as sapi_get_apireferral_ifnewuser_mock:
        results = await exchange.is_valid_account()
        assert results[0] is False
        assert isinstance(results[1], str)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value={})) as sapi_get_apireferral_ifnewuser_mock:
        results = await exchange.is_valid_account()
        assert results[0] is False
        assert isinstance(results[1], str)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value={"rebateWorking": False, "ifNewUser": True})) \
            as sapi_get_apireferral_ifnewuser_mock:
        results = await exchange.is_valid_account()
        assert results[0] is False
        assert isinstance(results[1], str)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value=None)) \
            as sapi_get_apireferral_ifnewuser_mock:
        results = await exchange.is_valid_account()
        assert results[0] is False
        assert isinstance(results[1], str)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(return_value={"rebateWorking": True, "ifNewUser": True})) \
            as sapi_get_apireferral_ifnewuser_mock:
        assert (await exchange.is_valid_account()) == (True, None)
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)
    with mock.patch.object(exchange._exchange.connector.client, "sapi_get_apireferral_ifnewuser",
                           mock.AsyncMock(side_effect=ccxt.async_support.InvalidNonce())) \
            as sapi_get_apireferral_ifnewuser_mock:
        with pytest.raises(trading_backend.TimeSyncError):
            await exchange.is_valid_account()
        sapi_get_apireferral_ifnewuser_mock.assert_called_once_with(params=params)



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import trading_backend.exchanges as exchanges


class FTX(exchanges.Exchange):
    SPOT_ID = "Octobot"
    MARGIN_ID = "Octobot"
    FUTURE_ID = "Octobot"
    REF_ID = "33613187"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'ftx'

    def get_orders_parameters(self, params=None) -> dict:
        params = super().get_orders_parameters(params)
        params.update({'externalReferralProgram': self._get_id()})
        return params

    async def _inner_is_valid_account(self) -> (bool, str):
        """
        Response doc from https://help.ftx.com/hc/en-us/articles/360044373831-External-Referral-Programs
        {
        'enabled': True,
        'externalReferralProgram': 'TradeX100',
        'readOnly': False,
        'requireWhitelistedIp': False,
        'subaccountNickname': None,
        'withdrawalEnabled': False
        }
        """
        # Seems to be unnecessary because the program name is overwritten in `get_orders_parameters`
        return await super()._inner_is_valid_account()



#  Drakkar-Software trading-backend
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import trading_backend.exchanges as exchanges


class Ascendex(exchanges.Exchange):
    SPOT_ID = "OctoBot"
    MARGIN_ID = "OctoBot"
    FUTURE_ID = "OctoBot"
    IS_SPONSORING = True

    @classmethod
    def get_name(cls):
        return 'ascendex'

    def _generate_order_id(self):
        new_id = "".join(self._exchange.connector.client.uuid().split("-"))
        return f"{self._get_id()}{new_id[len(self._get_id()):]}"

    def get_orders_parameters(self, params=None) -> dict:
        params = super().get_orders_parameters(params)
        params.update({'id': self._generate_order_id()})
        return params
